#ifndef _DAHDI_TDMOX_H
#define _DAHDI_TDMOX_H

#include <linux/interrupt.h> /* tasklet */

#define DAHDI_TDMOX_VERSION 1
#define DAHDI_TDMOX_PATCHLEVEL 2
#define DAHDI_TDMOX_SUBLEVEL 4
#define DAHDI_TDMOX_VERSIONSTR __stringify(DAHDI_TDMOX_VERSION) "." \
		__stringify(DAHDI_TDMOX_PATCHLEVEL) "." \
		__stringify(DAHDI_TDMOX_SUBLEVEL)

#define DAHDI_DYNAMIC_MAX_CHANS 256
#define DAHDI_TDMOX_TASKLET_QUOTA 12	/* Will handle 250Hz timer from DAHDI */
#define DAHDI_TDMOX_RXFIFO 6

#define DAHDI_TDMOX_TVHIST 2

struct dahdi_tdmox_tv {
	ktime_t tv[DAHDI_TDMOX_TVHIST];	/* Requested to schedule */
	unsigned int min;
	unsigned int max;
};
struct dahdi_tdmox_tasklet_time {
	struct dahdi_tdmox_tv req;	/* Requested to schedule */
	struct dahdi_tdmox_tv sched;	/* Scheduled */
	struct dahdi_tdmox_tv run;	/* Runned */
	struct dahdi_tdmox_tv exec;	/* Executed operation */
};
struct dahdi_tdmox_tasklet_stat {
	unsigned int req;
	unsigned int sched;
	unsigned int run;
	unsigned int exec;
	unsigned int errors;
	struct dahdi_tdmox_tasklet_time time;
};

struct dahdi_tdmox_tasklet {
	struct tasklet_struct tlet;
	/* 	pending has 1 incrementor & 1 decrimentor.
		No need in lock or mutex. */	  
	atomic_t pending;
};

struct dahdi_tdmox_span {
	char driver[20];	/* Which low-level driver to use */
	char addr[40];		/* Destination address */
	int numchans;		/* Number of channels */
	int timing;		/* Timing source preference */
	int spanno;		/* Span number (filled in by DAHDI) */
	int span_id_sysfs;	/* Pavel. Will remove that struct later. */
};

struct dahdi_tdmox {
	char addr[40];
	char dname[20];
	int err;
	struct kref kref;
	long rxjif;
	unsigned short txcnt;
	unsigned short rxcnt;
	struct dahdi_device *ddev;
	struct dahdi_span span;
	struct dahdi_chan *chans[DAHDI_DYNAMIC_MAX_CHANS];
	struct dahdi_tdmox_driver *driver;
	void *pvt;
	int timing;
	int master;
	unsigned char *msgbuf;
	struct device *dev;
	struct kfifo *rxfifo;
	int rxfifo_size;        /* Not aligned size of FIFO, we actually use */
	spinlock_t rxfifo_lock; /* See connent below (spin_lock_init) */
	/* Statistic */
	unsigned int slip;
	unsigned int skip;
	unsigned int rxnuerr;
	struct dahdi_tdmox_tv rxtv;
	struct dahdi_tdmox_tv txtv;

	struct list_head list;
};

struct dahdi_tdmox_driver {
	/*! Driver name (e.g. Eth) */
	const char *name;

	/*! Driver description */
	const char *desc;

	/*! Create a new transmission pipe */
	int (*create)(struct dahdi_tdmox *d, const char *address);

	/*! Destroy a created transmission pipe */
	void (*destroy)(struct dahdi_tdmox *d);

	/*! Transmit a given message */
	void (*transmit)(struct dahdi_tdmox *d, u8 *msg, size_t msglen);

	/*! Flush any pending messages */
	int (*flush)(void);

	struct list_head list;
	struct module *owner;

	/*! Numberic id of next device created by this driver. */
	unsigned int id;
};

struct dahdi_tdmox_ops {
	struct module *owner;
	int (*ioctl)(unsigned int cmd, unsigned long data);
};

int dahdi_tdmox_register_driver(struct dahdi_tdmox_driver *);
void dahdi_tdmox_unregister_driver(struct dahdi_tdmox_driver *);
void dahdi_tdmox_receive(struct dahdi_span *, unsigned char *, int);
int dahdi_tdmox_create(struct dahdi_tdmox_span *);
int dahdi_tdmox_destroy(struct dahdi_tdmox_span *);

/*== SECTION: kernel abstraction ==*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
	#define tdmox_strscpy(dst,src,len) strscpy(dst,src,len)
#else
	static inline size_t tdmox_strscpy(char *dst, const char *src, size_t len) {
		/* strncpy will zero tail of dst. I don't performance much, at least it's safer than strlcpy */
		if (len > 0) {
			strncpy(dst, src, len-1);
			dst[len] = '\0';
		}
		return 0; //Should return strlen(dst), but I don't use return value.
	}
#endif //KERNEL_VERSION(6, 8, 0)

#endif /* _DAHDI_TDMOX_H */
